/*
 * Copyright (C) 1996 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#define KEYMAP_C

#include "mutt.h"
#include "mutt_curses.h"
#include "keymap.h"

#include <string.h>
#include <ctype.h>

/* contains the last key the user pressed */
int LastKey;

struct keymap_t *Keymaps[MENU_MAX];

static struct keymap_t *findkey (struct keymap_t *k, int c)
{
  while (k)
  {
    if (k->key == c)
      return k;
    k = k->next;
  }
  return NULL;
}

static struct keymap_t *allocKey (void)
{
  struct keymap_t *p = (struct keymap_t *)safe_malloc(sizeof(struct keymap_t));

  memset(p, 0, sizeof(struct keymap_t));
  return p;
}

int parsekey (char **s)
{
  char *p = *s;

  if (*p == '\\')
  {
    p++;
    if (tolower (*p) == 'c') /* control char */
    {
      *s += 3;
      return (toupper (*(p+1)) - '@');
    }
    else if (*p == 'n')
    {
      *s += 2;
      return '\n';
    }
    else if (*p == 'r')
    {
      *s += 2;
      return '\r';
    }
    else if (*p == 't')
    {
      *s += 2;
      return '\t';
    }
    else if (*p == 'e')
    {
      *s += 2;
      return '\033';
    }
    else
    {
      *s += 2;
      return (*p);
    }
  }
  else if (strcasecmp ("pageup", p) == 0)
  {
    (*s) += 6;
    return KEY_PPAGE;
  }
  else if (strcasecmp ("pagedown", p) == 0)
  {
    *s += 8;
    return KEY_NPAGE;
  }
  else if (strcasecmp ("up", p) == 0)
  {
    *s += 2;
    return KEY_UP;
  }
  else if (strcasecmp ("down", p) == 0)
  {
    *s += 4;
    return KEY_DOWN;
  }
  else if (strcasecmp ("right", p) == 0)
  {
    *s += 5;
    return KEY_RIGHT;
  }
  else if (strcasecmp ("left", p) == 0)
  {
    *s += 4;
    return KEY_LEFT;
  }
  else if (strcasecmp ("delete", p) == 0)
  {
    *s += 6;
    return KEY_DC;
  }
  else if (strcasecmp ("backspace", p) == 0)
  {
    *s += 9;
    return KEY_BACKSPACE;
  }
  else if (!strcasecmp ("home", p))
  {
    *s += 4;
    return KEY_HOME;
  }
  else if (!strcasecmp ("end", p))
  {
    *s += 3;
    return KEY_END;
  }
  else
  {
    (*s)++;
    return (*p);
  }
}

static struct keymap_t *genkey (char **s)
{
  struct keymap_t *r = NULL;
  struct keymap_t *tmp = NULL;

  while (**s)
  {
    if (r)
    {
      tmp->seq = allocKey();
      tmp = tmp->seq;
    }
    else
      r = tmp = allocKey();
    tmp->key = parsekey(s);
  }
  return r;
}

void bindkey (char *s, int menu, int op, char *macro)
{
  struct keymap_t *top = Keymaps[menu];
  struct keymap_t *tmp;
  int key;

  if (!top)
  {
    Keymaps[menu] = tmp = genkey(&s);
    while (tmp->seq) tmp = tmp->seq;
    tmp->op = op;
    tmp->macro = safe_strdup(macro);
    return;
  }

  while (*s)
  {
    key = parsekey (&s);
    tmp = findkey (top, key);
    if (!tmp)
    {
      /* the rest of the sequence needs to be allocated */
      while (top->next) top = top->next;
      top->next = allocKey ();
      top = top->next;
      top->key = key;
      top->seq = genkey (&s);
      while (top->seq) top = top->seq;
      top->op = op;
      top->macro = safe_strdup (macro);
      return;
    }
    if (*s)
    {
      if (!tmp->seq)
      {
	if (tmp->op != OP_NULL)
	  return; /* looks like the user already bound the prefix! */

	/* just a null-op, so generate the rest of the sequence */
	tmp->seq = genkey (&s);
	tmp = tmp->seq;
	while (tmp->seq) tmp = tmp->seq;
	tmp->op = op;
	tmp->macro = safe_strdup (macro);
	return;
      }
      top = tmp->seq;
    }
    else
    {
      /* override the previous value */
      tmp->op = op;
      safe_free ((void **)&tmp->macro);
      tmp->macro = safe_strdup (macro);
      return;
    }
  }
}

static struct keymap_t *lookupkey (int key, struct keymap_t *map)
{
  while (map)
  {
    if (map->key == key)
      return map;
    map = map->next;
  }
  return NULL;
}

static void push_string (char *s)
{
  char *p = s + strlen (s) - 1;

  while (p >= s) ungetch (*p--);
}

int dokey (int menu)
{
  struct keymap_t *map = Keymaps[menu];

  FOREVER
  {
    if ((LastKey = ci_getch ()) == ERR)
      return (-1);
    map = lookupkey (LastKey, map);
    if (!map)
      return OP_NULL;
    if (!map->seq)
    {
      if (map->op == OP_MACRO)
      {
	push_string (map->macro);
	map = Keymaps[menu];
	continue;
      }
      return map->op;
    }
    map = map->seq;
  }
}

static void generic_bindings (int menu)
{
  bindkey ("?",  menu, OP_GENERIC_HELP, NULL);
  bindkey ("\\n", menu, OP_GENERIC_SELECT_ENTRY, NULL);
  bindkey ("\\r", menu, OP_GENERIC_SELECT_ENTRY, NULL);
  bindkey ("j",  menu, OP_GENERIC_NEXT_ENTRY, NULL);
  bindkey ("k",  menu, OP_GENERIC_PREVIOUS_ENTRY, NULL);
  bindkey ("h",  menu, OP_GENERIC_PREVIOUS_PAGE, NULL);
  bindkey ("l",  menu, OP_GENERIC_NEXT_PAGE, NULL);
  bindkey ("q",  menu, OP_GENERIC_EXIT, NULL);
  bindkey ("x",  menu, OP_GENERIC_EXIT, NULL);
  bindkey ("=",  menu, OP_GENERIC_FIRST_ENTRY, NULL);
  bindkey ("*",  menu, OP_GENERIC_LAST_ENTRY, NULL);
  bindkey ("\\cL", menu, OP_GENERIC_REFRESH, NULL);
  bindkey ("H",  menu, OP_GENERIC_TOP_OF_PAGE, NULL);
  bindkey ("M",  menu, OP_GENERIC_MIDDLE_OF_PAGE, NULL);
  bindkey ("L",  menu, OP_GENERIC_BOTTOM_OF_PAGE, NULL);

  bindkey ("pagedown", menu, OP_GENERIC_NEXT_PAGE, NULL);
  bindkey ("right", menu, OP_GENERIC_NEXT_PAGE, NULL);
  bindkey ("pageup", menu, OP_GENERIC_PREVIOUS_PAGE, NULL);
  bindkey ("left",  menu, OP_GENERIC_PREVIOUS_PAGE, NULL);
  bindkey ("up",    menu, OP_GENERIC_PREVIOUS_ENTRY, NULL);
  bindkey ("down",  menu, OP_GENERIC_NEXT_ENTRY, NULL);
}

void keymap_init (void)
{
  memset (Keymaps, 0, sizeof (struct keymap_t *) * MENU_MAX);

  bindkey ("a",    MENU_MAIN, OP_MAIN_ALIAS, NULL);
  bindkey ("B",    MENU_MAIN, OP_MAIN_BOTTOM_OF_PAGE, NULL);
  bindkey ("b",    MENU_MAIN, OP_MAIN_BOUNCE_MESSAGE, NULL);
  bindkey ("C",    MENU_MAIN, OP_MAIN_COPY_MESSAGE, NULL);
  bindkey ("c",    MENU_MAIN, OP_MAIN_CHANGE_FOLDER, NULL);
  bindkey ("D",    MENU_MAIN, OP_MAIN_DELETE_PATTERN, NULL);
  bindkey ("\\cD", MENU_MAIN, OP_MAIN_DELETE_THREAD, NULL);
  bindkey ("d",    MENU_MAIN, OP_MAIN_DELETE_MESSAGE, NULL);
  /* Decode option is added - pms@romance.kaist.ac.kr */
  bindkey ("e",    MENU_MAIN, OP_MAIN_DECODE_SAVE_MESSAGE, NULL);
  bindkey ("E",    MENU_MAIN, OP_MAIN_DECODE_COPY_MESSAGE, NULL);
  bindkey ("j",    MENU_MAIN, OP_MAIN_NEXT_UNDELETED, NULL);
  bindkey ("J",    MENU_MAIN, OP_MAIN_NEXT_MESSAGE, NULL);
#ifdef _PGPPATH
  bindkey ("\\cF", MENU_MAIN, OP_MAIN_FORGET_PASSPHRASE, NULL);
#endif /* _PGPPATH */
  bindkey ("F",    MENU_MAIN, OP_MAIN_FLAG_MESSAGE, NULL);
  bindkey ("f",    MENU_MAIN, OP_MAIN_FORWARD_MESSAGE, NULL);
#ifdef USE_POP
  bindkey ("G",    MENU_MAIN, OP_MAIN_FETCH_MAIL, NULL);
#endif /* USE_POP */
  bindkey ("g",    MENU_MAIN, OP_MAIN_GROUP_REPLY, NULL);
  bindkey ("H",    MENU_MAIN, OP_MAIN_TOP_OF_PAGE, NULL);
  bindkey ("h",    MENU_MAIN, OP_MAIN_DISPLAY_HEADERS, NULL);
  bindkey ("k",    MENU_MAIN, OP_MAIN_PREVIOUS_UNDELETED, NULL);
  bindkey ("K",    MENU_MAIN, OP_MAIN_PREVIOUS_MESSAGE, NULL);
  bindkey ("\\cL", MENU_MAIN, OP_MAIN_REFRESH, NULL);
  bindkey ("L",    MENU_MAIN, OP_MAIN_LIST_REPLY, NULL);
  bindkey ("l",    MENU_MAIN, OP_MAIN_LIMIT, NULL);
  bindkey ("M",    MENU_MAIN, OP_MAIN_MIDDLE_OF_PAGE, NULL);
  bindkey ("m",    MENU_MAIN, OP_MAIN_MAIL, NULL);
  bindkey ("N",    MENU_MAIN, OP_MAIN_MARK_AS_NEW, NULL);
  bindkey ("n",    MENU_MAIN, OP_MAIN_NEXT_UNDELETED, NULL);
  bindkey ("\\cN", MENU_MAIN, OP_MAIN_NEXT_THREAD, NULL);
  bindkey ("o",    MENU_MAIN, OP_MAIN_SORT_MESSAGES, NULL);
  bindkey ("p",    MENU_MAIN, OP_MAIN_PRINT_MESSAGE, NULL);
  bindkey ("\\cP", MENU_MAIN, OP_MAIN_PREVIOUS_THREAD, NULL);
  bindkey ("Q",    MENU_MAIN, OP_MAIN_QUIT, NULL);
  bindkey ("q",    MENU_MAIN, OP_MAIN_QUIT, NULL);
  bindkey ("r",    MENU_MAIN, OP_MAIN_REPLY, NULL);
  bindkey ("R",    MENU_MAIN, OP_MAIN_RECALL_MESSAGE, NULL);
  bindkey ("s",    MENU_MAIN, OP_MAIN_SAVE_MESSAGE, NULL);
  bindkey ("S",    MENU_MAIN, OP_MAIN_SEARCH_BODY, NULL);
  bindkey ("T",    MENU_MAIN, OP_MAIN_TAG_PATTERN, NULL);
  bindkey ("t",    MENU_MAIN, OP_MAIN_TAG_MESSAGE, NULL);
  bindkey ("\\cT", MENU_MAIN, OP_MAIN_UNTAG_PATTERN, NULL);
  bindkey ("U",    MENU_MAIN, OP_MAIN_UNDELETE_PATTERN, NULL);
  bindkey ("u",    MENU_MAIN, OP_MAIN_UNDELETE_MESSAGE, NULL);
  bindkey ("\\cU", MENU_MAIN, OP_MAIN_BROWSE_URL, NULL);
  bindkey ("v",    MENU_MAIN, OP_MAIN_VIEW_ATTACHMENTS, NULL);
  bindkey ("x",    MENU_MAIN, OP_MAIN_EXIT, NULL);
  bindkey ("z",    MENU_MAIN, OP_MAIN_NEXT_PAGE, NULL);
  bindkey ("Z",    MENU_MAIN, OP_MAIN_PREVIOUS_PAGE, NULL);
  bindkey ("*",    MENU_MAIN, OP_MAIN_LAST_MESSAGE, NULL);
  bindkey ("=",    MENU_MAIN, OP_MAIN_FIRST_MESSAGE, NULL);
  bindkey (" ",    MENU_MAIN, OP_MAIN_DISPLAY_MESSAGE, NULL);
  bindkey ("\\n",  MENU_MAIN, OP_MAIN_DISPLAY_MESSAGE, NULL);
  bindkey ("\\r", MENU_MAIN, OP_MAIN_DISPLAY_MESSAGE, NULL);
  bindkey ("$",    MENU_MAIN, OP_MAIN_SYNC_FOLDER, NULL);
  bindkey ("/",    MENU_MAIN, OP_MAIN_SEARCH, NULL);
  bindkey ("\\e/", MENU_MAIN, OP_MAIN_SEARCH_REVERSE, NULL);
  bindkey ("|",    MENU_MAIN, OP_MAIN_PIPE_MESSAGE, NULL);
  bindkey ("?",    MENU_MAIN, OP_MAIN_HELP, NULL);
  bindkey (":",    MENU_MAIN, OP_MAIN_ENTER_COMMAND, NULL);
  bindkey ("!",    MENU_MAIN, OP_MAIN_SUBSHELL, NULL);
  bindkey ("\\t",  MENU_MAIN, OP_MAIN_NEXT_NEW, NULL);
  bindkey ("%",    MENU_MAIN, OP_MAIN_DISPLAY_ADDRESS, NULL);
  bindkey (";",    MENU_MAIN, OP_MAIN_TAG_PREFIX, NULL);
   
  bindkey ("1", MENU_MAIN, OP_MAIN_JUMP, NULL);
  bindkey ("2", MENU_MAIN, OP_MAIN_JUMP, NULL);
  bindkey ("3", MENU_MAIN, OP_MAIN_JUMP, NULL);
  bindkey ("4", MENU_MAIN, OP_MAIN_JUMP, NULL);
  bindkey ("5", MENU_MAIN, OP_MAIN_JUMP, NULL);
  bindkey ("6", MENU_MAIN, OP_MAIN_JUMP, NULL);
  bindkey ("7", MENU_MAIN, OP_MAIN_JUMP, NULL);
  bindkey ("8", MENU_MAIN, OP_MAIN_JUMP, NULL);
  bindkey ("9", MENU_MAIN, OP_MAIN_JUMP, NULL);
  
  bindkey ("right", MENU_MAIN, OP_MAIN_NEXT_PAGE, NULL);
  bindkey ("pagedown", MENU_MAIN, OP_MAIN_NEXT_PAGE, NULL);
  bindkey ("left",  MENU_MAIN, OP_MAIN_PREVIOUS_PAGE, NULL);
  bindkey ("pageup", MENU_MAIN, OP_MAIN_PREVIOUS_PAGE, NULL);
  bindkey ("up",    MENU_MAIN, OP_MAIN_PREVIOUS_UNDELETED, NULL);
  bindkey ("down",  MENU_MAIN, OP_MAIN_NEXT_UNDELETED, NULL);

  /* defaults for the internal pager. */
  
  bindkey (" ",    MENU_PAGER, OP_PAGER_NEXT_PAGE, NULL);
  bindkey ("-",    MENU_PAGER, OP_PAGER_PREVIOUS_PAGE, NULL);
  bindkey ("\\n",  MENU_PAGER, OP_PAGER_NEXT_LINE, NULL);
  bindkey ("\\r",  MENU_PAGER, OP_PAGER_NEXT_LINE, NULL);
  bindkey ("^",    MENU_PAGER, OP_PAGER_TOP, NULL);
  bindkey ("/",    MENU_PAGER, OP_PAGER_SEARCH, NULL);
  bindkey ("?",    MENU_PAGER, OP_PAGER_HELP, NULL);
  bindkey ("$",    MENU_PAGER, OP_PAGER_BOTTOM, NULL);
  bindkey ("T",    MENU_PAGER, OP_PAGER_HIDE_QUOTED, NULL);
  bindkey ("\\cP", MENU_PAGER, OP_PAGER_HALF_UP, NULL);
  bindkey ("\\cN", MENU_PAGER, OP_PAGER_HALF_DOWN, NULL);
  bindkey ("\\cL", MENU_PAGER, OP_PAGER_REDRAW, NULL);
  bindkey ("backspace", MENU_PAGER, OP_PAGER_PREVIOUS_LINE, NULL);
  
  bindkey ("\\t", MENU_PAGER, OP_MAIN_NEXT_NEW, NULL);
  bindkey ("b", MENU_PAGER, OP_MAIN_BOUNCE_MESSAGE, NULL);
  bindkey ("c", MENU_PAGER, OP_MAIN_CHANGE_FOLDER, NULL);
  bindkey ("C", MENU_PAGER, OP_MAIN_COPY_MESSAGE, NULL);
  bindkey ("d", MENU_PAGER, OP_MAIN_DELETE_MESSAGE, NULL);
  bindkey ("\\cD", MENU_PAGER, OP_MAIN_DELETE_THREAD, NULL);
  bindkey ("F", MENU_PAGER, OP_MAIN_FLAG_MESSAGE, NULL);
  bindkey ("f", MENU_PAGER, OP_MAIN_FORWARD_MESSAGE, NULL);
  bindkey ("g", MENU_PAGER, OP_MAIN_GROUP_REPLY, NULL);
  bindkey ("h", MENU_PAGER, OP_MAIN_DISPLAY_HEADERS, NULL);
  bindkey ("i", MENU_PAGER, OP_MAIN_EXIT, NULL);
  bindkey ("j", MENU_PAGER, OP_MAIN_NEXT_UNDELETED, NULL);
  bindkey ("J", MENU_PAGER, OP_MAIN_NEXT_MESSAGE, NULL);
  bindkey ("k", MENU_PAGER, OP_MAIN_PREVIOUS_UNDELETED, NULL);
  bindkey ("K", MENU_PAGER, OP_MAIN_PREVIOUS_MESSAGE, NULL);
  bindkey ("L", MENU_PAGER, OP_MAIN_LIST_REPLY, NULL);
  bindkey ("m", MENU_PAGER, OP_MAIN_MAIL, NULL);
  bindkey ("n", MENU_PAGER, OP_MAIN_NEXT_UNDELETED, NULL);
  bindkey ("N", MENU_PAGER, OP_MAIN_MARK_AS_NEW, NULL);
  bindkey ("\\cN", MENU_PAGER, OP_MAIN_NEXT_THREAD, NULL);
  bindkey ("p", MENU_PAGER, OP_MAIN_PRINT_MESSAGE, NULL);
  bindkey ("\\cP", MENU_PAGER, OP_MAIN_PREVIOUS_THREAD, NULL);
  bindkey ("q", MENU_PAGER, OP_MAIN_EXIT, NULL);
  bindkey ("r", MENU_PAGER, OP_MAIN_REPLY, NULL);
  bindkey ("s", MENU_PAGER, OP_MAIN_SAVE_MESSAGE, NULL);
  bindkey ("t", MENU_PAGER, OP_MAIN_TAG_MESSAGE, NULL);
  bindkey ("u", MENU_PAGER, OP_MAIN_UNDELETE_MESSAGE, NULL);
  bindkey ("v", MENU_PAGER, OP_MAIN_VIEW_ATTACHMENTS, NULL);
  bindkey ("\\cU", MENU_PAGER, OP_MAIN_BROWSE_URL, NULL);
  bindkey ("x", MENU_PAGER, OP_MAIN_EXIT, NULL);
  bindkey ("|", MENU_PAGER, OP_MAIN_PIPE_MESSAGE, NULL);
  bindkey (":", MENU_PAGER, OP_MAIN_ENTER_COMMAND, NULL);

  bindkey ("pagedown", MENU_PAGER, OP_PAGER_NEXT_PAGE, NULL);
  bindkey ("pageup", MENU_PAGER, OP_PAGER_PREVIOUS_PAGE, NULL);
  bindkey ("up", MENU_PAGER, OP_MAIN_PREVIOUS_UNDELETED, NULL);
  bindkey ("right", MENU_PAGER, OP_MAIN_NEXT_UNDELETED, NULL);
  bindkey ("down", MENU_PAGER, OP_MAIN_NEXT_UNDELETED, NULL);
  bindkey ("left", MENU_PAGER, OP_MAIN_PREVIOUS_UNDELETED, NULL);

  /* defaults for the attachment menu. */

  generic_bindings (MENU_ATTACH);
  bindkey ("s", MENU_ATTACH, OP_ATTACH_SAVE_ENTRY, NULL);
  bindkey ("p", MENU_ATTACH, OP_ATTACH_PRINT_ENTRY, NULL);
  bindkey ("|", MENU_ATTACH, OP_ATTACH_PIPE_ENTRY, NULL);

  /* defaults for the send menu */

  bindkey ("?",  MENU_SEND, OP_SEND_HELP, NULL);
  bindkey (":",  MENU_SEND, OP_SEND_ENTER_COMMAND, NULL);
  bindkey ("a",  MENU_SEND, OP_SEND_ATTACH_FILE, NULL);
  bindkey ("b",  MENU_SEND, OP_SEND_EDIT_BCC, NULL);
  bindkey ("c",  MENU_SEND, OP_SEND_EDIT_CC, NULL);
  bindkey ("e",  MENU_SEND, OP_SEND_EDIT_MESSAGE, NULL);
  bindkey ("E",  MENU_SEND, OP_SEND_EDIT_HEADERS, NULL);
  bindkey ("f",  MENU_SEND, OP_SEND_EDIT_FCC, NULL);
  bindkey ("t",  MENU_SEND, OP_SEND_EDIT_TO, NULL);
  bindkey ("s",  MENU_SEND, OP_SEND_EDIT_SUBJECT, NULL);
  bindkey ("\\n", MENU_SEND, OP_SEND_SEND_MESSAGE, NULL);
  bindkey ("\\r", MENU_SEND, OP_SEND_SEND_MESSAGE, NULL);
  bindkey ("y",  MENU_SEND, OP_SEND_SEND_MESSAGE, NULL);
  bindkey ("n",  MENU_SEND, OP_SEND_ABORT_MESSAGE, NULL);
  bindkey ("P",  MENU_SEND, OP_SEND_POSTPONE_MESSAGE, NULL);
#ifdef _PGPPATH
  bindkey ("p",  MENU_SEND, OP_SEND_PGP_MENU, NULL);
#endif
#ifdef ISPELL
  bindkey ("i",  MENU_SEND, OP_SEND_ISPELL, NULL);
#endif

  generic_bindings (MENU_POST);
  bindkey ("d", MENU_POST, OP_GENERIC_DELETE_ENTRY, NULL);
  bindkey ("u", MENU_POST, OP_GENERIC_UNDELETE_ENTRY, NULL);

  generic_bindings (MENU_SENDATTACH);
  bindkey ("d", MENU_SENDATTACH, OP_GENERIC_DELETE_ENTRY, NULL);
  bindkey ("a", MENU_SENDATTACH, OP_SENDATTACH_ADD_ATTACHMENT, NULL);
  bindkey ("s", MENU_SENDATTACH, OP_SENDATTACH_ENTER_DESCRIPTION, NULL);

  /* These menus only use the generic keybindings. */
  generic_bindings (MENU_FOLDER);
  generic_bindings (MENU_ALIAS);
  generic_bindings (MENU_URL);
}
