/*
 * Copyright (C) 1996 Michael R. Elkins <me@cs.hmc.edu>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"
#include "keymap.h"

#include <string.h>

#define OFFSET 1

MUTTMENU *mutt_newMenu (const char *title)
{
  MUTTMENU *p = (MUTTMENU *)safe_malloc(sizeof(MUTTMENU));

  memset (p, 0, sizeof(MUTTMENU));
  p->title = safe_strdup(title);
  p->redraw = REDRAW_FULL;
  return (p);
}

void mutt_menuDestroy (MUTTMENU **p)
{
  safe_free ((void **)&(*p)->title);
  safe_free ((void **)p);
}

int mutt_menuLoop (MUTTMENU *menu)
{
  int i;
  char buffer[STRING];

  FOREVER
  {
    /*
     * See if all or part of the screen needs to be updated.
     */
    switch (menu->redraw) {
    case REDRAW_FULL:
      clear ();
      SETCOLOR (MT_COLOR_STATUS);
      mvprintw (LINES-2, 0, "%-*.*s", COLS, COLS, "");
      mvaddstr (LINES-2, 0, menu->title);
      SETCOLOR (MT_COLOR_NORMAL);
      menu->pagelen = LINES-3;

      /* fall through to REDRAW_INDEX */

    case REDRAW_INDEX:
      for (i = menu->top; i < menu->top + menu->pagelen; i++)
      {
	if (i < menu->max)
	{
	  menu->makeEntry (buffer, i, sizeof(buffer));

	  CLEARLINE (i - menu->top + OFFSET);

	  if (option (OPTARROWCURSOR))
	  {
	    if (i == menu->current)
	    {
	      SETCOLOR (MT_COLOR_INDICATOR);
	      addstr ("->");
	      SETCOLOR (MT_COLOR_NORMAL);
	    }
	    buffer[COLS-3] = 0;
	    move (i - menu->top + OFFSET, 3);
	    addstr (buffer);
	  }
	  else
	  {
	    buffer[COLS-1] = 0;
	    
	    if (i == menu->current)
	      SETCOLOR (MT_COLOR_INDICATOR);
	    printw ("%-*.*s", COLS, COLS, buffer);
	    if ( i == menu->current)
	      SETCOLOR (MT_COLOR_NORMAL);
	  }
	}
	else
	  CLEARLINE (i-menu->top+OFFSET);
      }
      menu->redraw = 0;
      break;

    case REDRAW_MOTION:
      move (menu->oldcurrent + OFFSET - menu->top, 0);
      if (option (OPTARROWCURSOR))
      {
	/* clear the pointer */
	addstr ("  ");

	/* now draw it in the new location */
	move (menu->current + OFFSET - menu->top, 0);
	SETCOLOR (MT_COLOR_INDICATOR);
	addstr ("->");
	SETCOLOR (MT_COLOR_NORMAL);
      }
      else
      {
	/* erase the current indicator */
	clrtoeol ();
	menu->makeEntry (buffer, menu->oldcurrent, sizeof (buffer));
	buffer[COLS-1] = 0;
	printw ("%-*.*s", COLS, COLS, buffer);

	/* now draw the new one to reflect the change */
	menu->makeEntry (buffer, menu->current, sizeof (buffer));
	buffer[COLS-1] = 0;
	move (menu->current - menu->top + OFFSET, 0);
	SETCOLOR (MT_COLOR_INDICATOR);
	printw ("%-*.*s", COLS, COLS, buffer);
	SETCOLOR (MT_COLOR_NORMAL);
      }
      menu->redraw = 0;
      break;

    case REDRAW_CURRENT:
      move (menu->current + OFFSET - menu->top, 0);
      clrtoeol ();
      if (option (OPTARROWCURSOR))
      {
	SETCOLOR (MT_COLOR_INDICATOR);
	addstr ("->");
	SETCOLOR (MT_COLOR_NORMAL);
	menu->makeEntry (buffer, menu->current, sizeof (buffer));
	buffer[LINES-3] = 0;
	addstr (buffer);
      }
      else
      {
	buffer[COLS-1] = 0;
	SETCOLOR (MT_COLOR_INDICATOR);
	printw ("%-*.*s", COLS, COLS, buffer);
	SETCOLOR (MT_COLOR_NORMAL);
      }

      menu->redraw = 0;
      break;
    }

    /* move the cursor out of the way */
    if (option (OPTARROWCURSOR))
    {
      move (menu->current - menu->top + OFFSET, 3);
    }
    else
    {
      move (menu->current - menu->top + OFFSET, COLS-1);
    }

    refresh ();
    i = dokey (menu->menu);
    ci_clear_error ();
    if (i == ERR)
    {
#ifdef USE_SLANG_CURSES
      if (Sigwinch)
      {
	mutt_resize_screen ();
	menu->redraw = REDRAW_FULL;
	Sigwinch = 0;
      }
#endif
      continue;
    }
    switch (i) {
    case OP_GENERIC_NEXT_ENTRY:
      if (menu->current < menu->max - 1)
      {
	menu->oldcurrent = menu->current;
	menu->current++;
	if (menu->current >= menu->top + menu->pagelen)
	{
	  menu->top = menu->current;
	  menu->redraw = REDRAW_INDEX;
	}
	else
	  menu->redraw = REDRAW_MOTION;
      }
      else
      {
	beep ();
	mutt_error ("You are on the last entry.");
      }
      break;

    case OP_GENERIC_PREVIOUS_ENTRY:
      if (menu->current > 0)
      {
	menu->oldcurrent = menu->current;
	if (--menu->current < menu->top)
	{
	  if ((menu->top = menu->current - menu->pagelen + 1) < 0)
	    menu->top = 0;
	  menu->redraw = REDRAW_INDEX;
	}
	else
	  menu->redraw = REDRAW_MOTION;
      }
      else
      {
	beep ();
	mutt_error("You are on the first entry.");
      }
      break;
	
    case OP_GENERIC_NEXT_PAGE:
      if (menu->top + menu->pagelen <= menu->max - 1)
      {
	menu->top += menu->pagelen;
	menu->current = menu->top;
	menu->redraw = REDRAW_INDEX;
      }
      else
      {
	beep ();
	mutt_error ("You are on the last page.");
      }
      break;

    case OP_GENERIC_PREVIOUS_PAGE:
      if (menu->top > 0)
      {
	if ((menu->top -= menu->pagelen) < 0)
	  menu->top = 0;
	menu->current = menu->top + menu->pagelen - 1;
	menu->redraw = REDRAW_INDEX;
      }
      else
      {
	beep ();
	mutt_error("You are on the first page.");
      }
      break;

    case OP_GENERIC_FIRST_ENTRY:
      menu->oldcurrent = menu->current;
      if (menu->top > 0)
	menu->redraw = REDRAW_INDEX;
      else
	menu->redraw = REDRAW_MOTION;
      menu->top = menu->current = 0;
      break;

    case OP_GENERIC_LAST_ENTRY:
      menu->oldcurrent = menu->current;
      menu->current = menu->max - 1;
      if (menu->current > menu->top + menu->pagelen - 1)
      {
	while (menu->top + menu->pagelen < menu->current)
	  menu->top += menu->pagelen;
	menu->redraw = REDRAW_INDEX;
      }
      else
	menu->redraw = REDRAW_MOTION;
      break;

    case OP_GENERIC_TOP_OF_PAGE:
      menu->oldcurrent = menu->current;
      menu->current = menu->top;
      menu->redraw = REDRAW_MOTION;
      break;

    case OP_GENERIC_MIDDLE_OF_PAGE:
      menu->oldcurrent = menu->current;
      i = menu->top + menu->pagelen;
      if (i > menu->max - 1)
	i = menu->max - 1;
      menu->current = menu->top + (i - menu->top) / 2;
      break;

    case OP_GENERIC_BOTTOM_OF_PAGE:
      menu->oldcurrent = menu->current;
      menu->current = menu->top + menu->pagelen - 1;
      if (menu->current > menu->max - 1)
	menu->current = menu->max - 1;
      menu->redraw = REDRAW_MOTION;
      break;

    case OP_GENERIC_REFRESH:
      clearok (stdscr, TRUE);
      menu->redraw = REDRAW_FULL;
      break;

    case OP_GENERIC_HELP:
      ci_help (menu->menu);
      menu->redraw = REDRAW_FULL;
      break;

    case OP_NULL:
      beep ();
      mutt_error ("Key is not bound.  Press '?' for help.");
      break;

    default:
      return (i);
    }
  }
  /* not reached */
}
