/*
 * Copyright (C) 1996 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

/* Attachment menu for the curses interface. */

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"
#include "keymap.h"

#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <string.h>

struct attachptr {
  BODY *content;
  struct attachptr *next;
};

struct attachptr **idx = NULL;

static struct attachptr *gen_attach_list (BODY *m)
{
  struct attachptr *top = NULL, *end = NULL, *new;

  while (m)
  {
    if (m->type == TYPEMULTIPART)
    {
      if (top)
        end->next = gen_attach_list(m->parts);
      else
        end = top = gen_attach_list(m->parts);
      while (end->next) end = end->next;
    }
    else
    {
      new = (struct attachptr *)safe_malloc(sizeof(struct attachptr));
      new->next = NULL;
      new->content = m;

      if (top)
      {
        end->next = new;
        end = new;
      }
      else
        top = end = new;

      if (m->type == TYPEMESSAGE && !strcasecmp(m->subtype, "rfc822") &&
	  is_multipart(m->parts))
      {
	end->next = gen_attach_list(m->parts);
	while (end->next) end = end->next;
      }
    }
    m = m->next;
  }
  return (top);
}

static int can_print (BODY *m)
{
  if (m->type == TYPETEXT)
  {
    if (strcasecmp("plain", m->subtype) == 0) return 1;
  }
  else if (m->type == TYPEAPPLICATION)
  {
    if (strcasecmp("postscript", m->subtype) == 0) return 1;
  }
  return 0;
}

static void display_attachment (BODY *a)
{
  STATE *s;
  char tempfile[_POSIX_PATH_MAX], descrip[SHORT_STRING];

  mutt_mktemp (tempfile);
#ifdef METAMAIL
  if (mutt_needs_metamail (a))
  {
    char buffer[STRING];

    mutt_save_attachment (a, tempfile);
    snprintf (buffer, sizeof (buffer),
	      "%s -b -c %s/%s -P -z %s", METAMAIL, TYPE(a->type),
	      a->subtype, tempfile);
    endwin ();
    mutt_system (buffer);
    return;
  }
#endif

  s = mutt_new_state ();
  s->fpout = fopen (tempfile, "w");
  s->fpin = Context->folder->fp;
  s->displaying = 1;
  mutt_body_handler (a, s);
  fclose (s->fpout);
  mutt_free_state (&s);
  
  if (a->description)
    strfcpy (descrip, a->description, sizeof (descrip));
  else
    snprintf (descrip, sizeof (descrip),
	      "Attachment: %s/%s", TYPE(a->type), a->subtype);

  mutt_do_pager (descrip, tempfile);
}

static void pipe_attachment (BODY *b, const char *path)
{
  STATE s;

  memset (&s, 0, sizeof (STATE));
  s.fpin = Context->folder->fp;
  endwin ();
  mutt_create_filter (path, &s.fpout, NULL);
  mutt_body_handler (b, &s);
  fclose (s.fpout);
  wait (NULL);
  ci_any_key_to_continue (NULL);
}

void attach_makeEntry (char *b, int num, size_t blen)
{
  char t[SHORT_STRING];
  BODY *m = idx[num]->content;

  snprintf(t, sizeof(t), "%s/%s", TYPE(m->type), m->subtype);
  snprintf(b, blen, " %2d %-20.20s  %-6.6s (%6ld) %s", num+1, t,
	  ENCODING(m->encoding), m->length,
	  m->description ? m->description : (m->filename ? m->filename : ""));
}

void ci_attach (BODY *cur)
{
  struct attachptr *list, *plist;
  int idxlen = 0, count = 0;
  char buffer[STRING];
  MUTTMENU *menu;

  plist = list = gen_attach_list(cur);
  for (count=0; plist; count++, plist = plist->next)
  {
    if (count == idxlen)
    {
      if (idxlen == 0)
        idx = safe_malloc(sizeof(struct attachptr *) * ((idxlen = 5)));
      else
        safe_realloc((void **)&idx, ((idxlen += 5)) * sizeof(struct attachptr *));
    }
    idx[count] = plist;
  }

  snprintf (buffer, sizeof (buffer), "---Mutt: Attachment Mode [%d attachment%s]", count, count != 1 ? "s" : "");
  menu = mutt_newMenu (buffer);
  menu->max = count;
  menu->makeEntry = attach_makeEntry;
  menu->menu = MENU_ATTACH;

  FOREVER
  {
    switch (mutt_menuLoop (menu)) {
    case OP_GENERIC_SELECT_ENTRY:
      display_attachment (idx[menu->current]->content);
      menu->redraw = REDRAW_FULL;
      break;

    case OP_ATTACH_PRINT_ENTRY:
      if (can_print (idx[menu->current]->content))
	pipe_attachment (idx[menu->current]->content, Print);
      else
      {
	beep ();
        mutt_error ("I don't know how to print that!");
      }
      break;

    case OP_ATTACH_PIPE_ENTRY:
      if (ci_get_field ("Pipe to: ", buffer, sizeof (buffer), 0) == 0 &&
	  buffer[0])
      {
	mutt_expand_path (buffer, sizeof (buffer));
	pipe_attachment (idx[menu->current]->content, buffer);
	ci_any_key_to_continue (NULL);
      }
      break;

    case OP_ATTACH_SAVE_ENTRY:
      if (idx[menu->current]->content->filename)
	strfcpy(buffer, idx[menu->current]->content->filename, sizeof(buffer));
      else
	buffer[0] = 0;
      if (ci_get_field ("Save to file: ", buffer, sizeof (buffer), M_FILE) != 0 ||
	  !buffer[0])
	break;
      mutt_expand_path (buffer, sizeof (buffer));
      if (mutt_save_attachment (idx[menu->current]->content, buffer) == 0)
	mutt_error ("Attachment saved.");
      else
      {
	beep ();
	mutt_error ("Error while saving attachment!");
      }
      break;

    case OP_GENERIC_EXIT:
      while (count-- > 0) safe_free((void **)&idx[count]);
      safe_free ((void **)&idx);
      mutt_menuDestroy (&menu);
      return;
    }
  }
  /* not reached */
}
